/*******************************************************************************
**		This DLL includes functions that allow Windows NT applications to       **
**		perform I/O operations both to physical memory and comm ports.          **
**                                                                            **
**    B&B Technologies Ltd. Co.                                               **
**    6610 Gulton Court NE, Suite A           Created by: Juan Ortiz-Moyet    **
**		Albuquerque, NM 87109-1477              Last modification:              **
**		(505) 345-9499                           Added support for PCI bus		**
**    Fax: (505)345-9699                       and EISA bus devices.				**
**		e-mail: bbtech@roadrunner.com				 Made on: July 16, 1998				**
**															 By: Juan Ortiz-Moyet            **
*******************************************************************************/

#include "ntsat.h"

	/* Queue of active session resources.	*/
static struct Table Session[MAX_SESSION_NUMBER];

//	/* Driver initialization flag.	*/
//static BOOL FirstSession;

	/* Interrupt session flag.	*/
static DWORD ID;

/*******************************************************************************
**	   DllEntryPoint():	Function to initialize and terminate DLL instance		**
**																										**
**		Arguments:  check Win32 function reference for details						**
**																										**
**		Return value:	Success = TRUE															**
**							Failure = FALSE							 							**
**										 (specific error code - GetLastError() )			**
*******************************************************************************/
BOOL WINAPI DllEntryPoint(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
 register DWORD i;

 switch (fdwReason)
   {
    case DLL_PROCESS_ATTACH:
      for ( i=0; i<MAX_SESSION_NUMBER; i++)  /* When the DLL is first loaded   */
        {                                    /* by any process, it must        */
         Session[i].Handle = NULL;          	/* clear the contents of the      */
         Session[i].Device = INVALID_DEVICE_NUMBER;
         Session[i].PhysMem1 = NULL;         /* Session Table for the code to  */
         Session[i].PhysMem2 = NULL;         /* execute without errors.			 */
			Session[i].Occurrence = -1;
         Session[i].WndHandle = NULL;
         Session[i].IsrEvent = NULL;
         Session[i].TerminateIsr = FALSE;
        }

//      	/* Signal initialization of driver session.	*/
//      FirstSession = TRUE;
      ID = 0;
      break;

    case DLL_PROCESS_DETACH:
         /* Stop the WinRT driver.  */
//      DriverControl(Stop_Driver);  Commented out to allow USER accounts
      break;                         //to call SAT functions...

    case DLL_THREAD_ATTACH:
    case DLL_THREAD_DETACH:
    default: break;
   }

 return(TRUE);
}


/*******************************************************************************
**	   SessionControl():	Function to initialize and terminate I/O session		**
**																										**
**		Arguments:  pdwSessionID - number that describes I/O session; if eTask	**
**						is Start_Session, it is equal to the device to be accessed.	**
**						eTask - type of operation to perform								**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD SessionControl(uInt32 *pdwSessionID, enum SCtrlFunc eTask)
{
 DWORD dwErrorCode = ERROR_SUCCESS;
 DWORD dwDeviceNumber;
 char szDeviceName[MAX_WINRT_NAME];
 char dummy[MAX_DEVICE_DIGITS];
 register int i;

 switch (eTask)
  {
   case (Start_Session):

//     if (FirstSession)
//     	 {
//        DriverControl(Start_Driver);  Commented out to allow NT USER accounts
//        FirstSession = FALSE;           to call SAT functions...
//       }

     dwDeviceNumber = *pdwSessionID;

 			/*	Search for an available index in Session table.	*/
     i = 1;					 /* Do not touch index 0; that one is always invalid.	*/
     while ( (Session[i].Device != INVALID_DEVICE_NUMBER) &&
 		  ( i < MAX_SESSION_NUMBER) ) i++;

     if ( Session[i].Device != INVALID_DEVICE_NUMBER )
     	 {
        *pdwSessionID = 0;       /* Invalidate SessionID.	*/
        return( 1 );      			/* Set error: no more devices are available.	*/
       }

         /* Save available index as new SessionID.	*/
     *pdwSessionID = i;

			/*	Generate DeviceName string with given device number.	*/
	  lstrcpy(szDeviceName, "\\\\.\\WinRTdev");
     itoa((int)(dwDeviceNumber), dummy, 10);
     lstrcat(szDeviceName, dummy);

			/*	Obtain handle to new session.	*/
     Session[*pdwSessionID].Handle = CreateFile( (LPCTSTR)(szDeviceName), 0,
     					FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);

     if ( Session[*pdwSessionID].Handle == INVALID_HANDLE_VALUE )
     	{                                        /* Check for invalid handle */
    	 Session[*pdwSessionID].Handle = NULL;
       Session[*pdwSessionID].Device = INVALID_DEVICE_NUMBER;
       *pdwSessionID = 0;                      /* Invalidate SessionID		*/
    	 return( GetLastError() );
   	}

     		/*	Update the Session table to reflect the new device.	*/
     Session[*pdwSessionID].Device = dwDeviceNumber;
     break;

   case (End_Session):

     /* Close handle to I/O device.	 */
     if ( !CloseHandle(Session[*pdwSessionID].Handle) )
       dwErrorCode = GetLastError();  				/* Catch error on API call. */

     Session[*pdwSessionID].Handle = NULL;
     Session[*pdwSessionID].Device = INVALID_DEVICE_NUMBER;

		/* If session started an IsrFunction thread, terminate the thread.	*/
     if (Session[*pdwSessionID].IsrEvent != NULL )
     	{
       Session[*pdwSessionID].TerminateIsr = TRUE;
       SetEvent(Session[*pdwSessionID].IsrEvent);
      }

      /* First element of Session table always contains invalid data.	*/
     *pdwSessionID = 0;			/* Invalidate SessionID		*/
     break;

   default: break;
  }

 return( dwErrorCode );
}


/*******************************************************************************
**	   ByteIo():	Function to execute byte memory I/O operations					**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulOffset - physical address to manage								**
**						ulNumBytes - number of BYTES to manage								**
**						pbData - address to hold requested byte(s)		 				**
**						iResGroup - especifies resource group for transfer				**
**									  ( TRUE => Secondary, FALSE => Primary )				**
**																										**
**		Return value:	ERROR_SUCCESS (See Winerror.h)									**
*******************************************************************************/
DWORD ByteIo(DWORD dwSessionID, BOOL iTask, ULONG ulOffset,
					ULONG ulNumBytes, PBYTE pbData, BOOL iResGroup)
{
 ULONG  i;
 PBYTE  pbAddress;

 if ( !iResGroup )
 	pbAddress = (PBYTE)(Session[dwSessionID].PhysMem1);
 else
   pbAddress = (PBYTE)(Session[dwSessionID].PhysMem2);

 if (pbAddress == NULL)              /* Check for invalid pointer before	*/
 	return ( ERROR_INVALID_HANDLE );  /* attempting any access.					*/

 pbAddress += ulOffset;     			 /* Start access at requested offset. 	*/

 if ( iTask )
  {
   	/* Read byte(s) into given location. 	*/
   for ( i = 0; i < ulNumBytes; i++ )
		*( pbData + i ) = *( pbAddress + i );
  }
 else
  {
   	/* Write byte(s) to specified location.	*/
   for ( i = 0; i < ulNumBytes; i++ )
		*( pbAddress + i ) = *( pbData + i );
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   WordIo():	Function to execute 16-bit memory I/O operations				**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulOffset - physical address to manage								**
**						ulNumBytes - number of BYTES to manage								**
**						pwData - address to hold requested word(s)		 				**
**						iResGroup - especifies resource group for transfer				**
**									  ( TRUE => Secondary, FALSE => Primary )				**
**																										**
**		Return value:	ERROR_SUCCESS (See Winerror.h)									**
*******************************************************************************/
DWORD WordIo(DWORD dwSessionID, BOOL iTask, ULONG ulOffset,
					ULONG ulNumBytes, PUSHORT pwData, BOOL iResGroup)
{
 ULONG	 i;
 PUSHORT	 pwAddress;

 if ( !iResGroup )
 	pwAddress = (PUSHORT)(Session[dwSessionID].PhysMem1);
 else
   pwAddress = (PUSHORT)(Session[dwSessionID].PhysMem2);

 if (pwAddress == NULL)              /* Check for invalid pointer before	*/
 	return ( ERROR_INVALID_HANDLE );  /* attempting any access.					*/

 pwAddress += ulOffset;     			 /* Start access at requested offset. 	*/

 if ( iTask )
  {
   	/* Read word(s) into given location. 	*/
   for ( i = 0; i < ulNumBytes/sizeof(short); i++ )
		*( pwData + i ) = *( pwAddress + i );
  }
 else
  {
   	/* Write word(s) to specified location.	*/
   for ( i = 0; i < ulNumBytes/sizeof(short); i++ )
		*( pwAddress + i ) = *( pwData + i );
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   DwordIo():	Function to execute 32-bit memory I/O operations				**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulOffset - physical address to manage								**
**						ulNumBytes - number of BYTES to manage								**
**						pulData - address to hold requested long(s)		 				**
**						iResGroup - especifies resource group for transfer				**
**									  ( TRUE => Secondary, FALSE => Primary )				**
**																										**
**		Return value:	ERROR_SUCCESS (See Winerror.h)									**
*******************************************************************************/
DWORD DwordIo(DWORD dwSessionID, BOOL iTask, ULONG ulOffset,
					ULONG ulNumBytes, PULONG pulData, BOOL iResGroup)
{
 ULONG	i;
 PULONG	pulAddress;

 if ( !iResGroup )
 	pulAddress = (PULONG)(Session[dwSessionID].PhysMem1);
 else
   pulAddress = (PULONG)(Session[dwSessionID].PhysMem2);

 if (pulAddress == NULL)              /* Check for invalid pointer before	*/
 	return ( ERROR_INVALID_HANDLE );  /* attempting any access.					*/

 pulAddress += ulOffset;     			 /* Start access at requested offset. 	*/

 if ( iTask )
  {
   	/* Read dword(s) into given location. 	*/
   for ( i = 0; i < ulNumBytes/sizeof(long); i++ )
		*( pulData + i ) = *( pulAddress + i );
  }
 else
  {
   	/* Write dword(s) to specified location.	*/
   for ( i = 0; i < ulNumBytes/sizeof(long); i++ )
		*( pulAddress + i ) = *( pulData + i );
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   SingleIo():	Function to execute single-precision memory I/O operations	**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulOffset - physical address to manage								**
**						ulNumBytes - number of BYTES to manage								**
**						pfData - address to hold requested single(s)		 				**
**						iResGroup - especifies resource group for transfer				**
**									  ( TRUE => Secondary, FALSE => Primary )				**
**																										**
**		Return value:	ERROR_SUCCESS (See Winerror.h)									**
*******************************************************************************/
DWORD SingleIo(DWORD dwSessionID, BOOL iTask, ULONG ulOffset,
					ULONG ulNumBytes, PFLOAT pfData, BOOL iResGroup)
{
 ULONG	 i;
 PFLOAT	 pfAddress;

 if ( !iResGroup )
 	pfAddress = (PFLOAT)(Session[dwSessionID].PhysMem1);
 else
   pfAddress = (PFLOAT)(Session[dwSessionID].PhysMem2);

 if (pfAddress == NULL)              /* Check for invalid pointer before	*/
 	return ( ERROR_INVALID_HANDLE );  /* attempting any access.					*/

 pfAddress += ulOffset;     			 /* Start access at requested offset. 	*/

 if ( iTask )
  {
   	/* Read single(s) into given location. 	*/
   for ( i = 0; i < ulNumBytes/sizeof(float); i++ )
		*( pfData + i ) = *( pfAddress + i );
  }
 else
  {
   	/* Write single(s) to specified location.	*/
   for ( i = 0; i < ulNumBytes/sizeof(float); i++ )
		*( pfAddress + i ) = *( pfData + i );
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   DoubleIo():	Function to execute double-precision memory I/O operations	**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulOffset - physical address to manage								**
**						ulNumBytes - number of BYTES to manage								**
**						pdData - address to hold requested double(s)		 				**
**						iResGroup - especifies resource group for transfer				**
**									  ( TRUE => Secondary, FALSE => Primary )				**
**																										**
**		Return value:	ERROR_SUCCESS (See Winerror.h)									**
*******************************************************************************/
DWORD DoubleIo(DWORD dwSessionID, BOOL iTask, ULONG ulOffset,
					ULONG ulNumBytes, PDOUBLE pdData, BOOL iResGroup)
{
 ULONG	 i;
 PDOUBLE	 pdAddress;

 if ( !iResGroup )
 	pdAddress = (PDOUBLE)(Session[dwSessionID].PhysMem1);
 else
   pdAddress = (PDOUBLE)(Session[dwSessionID].PhysMem2);

 if (pdAddress == NULL)              /* Check for invalid pointer before	*/
 	return ( ERROR_INVALID_HANDLE );  /* attempting any access.					*/

 pdAddress += ulOffset;     			 /* Start access at requested offset. 	*/

 if ( iTask )
  {
   	/* Read double(s) into given location. 	*/
   for ( i = 0; i < ulNumBytes/sizeof(double); i++ )
		*( pdData + i ) = *( pdAddress + i );
  }
 else
  {
   	/* Write double(s) to specified location.	*/
   for ( i = 0; i < ulNumBytes/sizeof(double); i++ )
		*( pdAddress + i ) = *( pdData + i );
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   PortIo():	Function to execute data transfers with 8-bit ports			**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulAddress - physical address to manage								**
**						ulNumElements - number of data elements to transfer			**
**						pbData - address that holds session buffer		 				**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
*******************************************************************************/
DWORD PortIo(DWORD dwSessionID, BOOL iTask, ULONG ulAddress,
					ULONG ulNumElements, PBYTE pbData)
{
 ULONG	ulCount;
 register i;

 if ( iTask )
  {
   WINRT_CONTROL_ITEM	ReadCommand[] = {{INP_BA, 0, 0}};
   for (i=0; i < ulNumElements; i++)
    {
     ReadCommand[0].port = ulAddress;

		/* Process input command and copy bytes to specified location.		*/
	  if ( !WinRTProcessIoBuffer(Session[dwSessionID].Handle, ReadCommand,
   										sizeof(ReadCommand), &ulCount) )
		return ( GetLastError() );

	  *(pbData + i) = (BYTE)(ReadCommand[0].value);
    }
  }
 else
  {
   WINRT_CONTROL_ITEM	WriteCommand[] = {{OUTP_BA, 0, 0}};
   for (i=0; i < ulNumElements; i++)
    {
	  WriteCommand[0].port = ulAddress;
     WriteCommand[0].value = (ULONG)( *(pbData + i));

			/* Process output command and write byte to specified location.		*/
	  if ( !WinRTProcessIoBuffer(Session[dwSessionID].Handle, WriteCommand,
   										sizeof(WriteCommand), &ulCount) )
		return ( GetLastError() );
    }
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   PortIo_16():	Function to execute data transfers with 16-bit ports		**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulAddress - physical address to manage								**
**						ulNumElements - number of data elements to transfer			**
**						pwData - address that holds session buffer		 				**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
*******************************************************************************/
DWORD PortIo_16(DWORD dwSessionID, BOOL iTask, ULONG ulAddress,
						ULONG ulNumElements, PUSHORT pwData)
{
 ULONG	ulCount;
 register i;

 if ( iTask )
  {
   WINRT_CONTROL_ITEM	ReadCommand[] = {{INP_WA, 0, 0}};
   for (i=0; i < ulNumElements; i++)
    {
     ReadCommand[0].port = ulAddress;

		/* Process input command and copy bytes to specified location.		*/
	  if ( !WinRTProcessIoBuffer(Session[dwSessionID].Handle, ReadCommand,
   										sizeof(ReadCommand), &ulCount) )
		return ( GetLastError() );

	  *(pwData + i) = (USHORT)(ReadCommand[0].value);
    }
  }
 else
  {
   WINRT_CONTROL_ITEM	WriteCommand[] = {{OUTP_WA, 0, 0}};
   for (i=0; i < ulNumElements; i++)
    {
	  WriteCommand[0].port = ulAddress;
     WriteCommand[0].value = (ULONG)( *(pwData + i));

			/* Process output command and write byte to specified location.		*/
	  if ( !WinRTProcessIoBuffer(Session[dwSessionID].Handle, WriteCommand,
   										sizeof(WriteCommand), &ulCount) )
		return ( GetLastError() );
    }
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   PortIo_32():	Function to execute data transfers with 32-bit ports		**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iTask - specifies operation to perform								**
**									  ( TRUE => Read, FALSE => Write )						**
**						ulAddress - physical address to manage								**
**						ulNumElements - number of data elements to transfer			**
**						pulData - address that holds session buffer		 				**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
*******************************************************************************/
DWORD PortIo_32(DWORD dwSessionID, BOOL iTask, ULONG ulAddress,
						ULONG ulNumElements, PULONG pulData)
{
 ULONG	ulCount;
 register i;

 if ( iTask )
  {
   WINRT_CONTROL_ITEM	ReadCommand[] = {{INP_LA, 0, 0}};
   for (i=0; i < ulNumElements; i++)
    {
     ReadCommand[0].port = ulAddress;

		/* Process input command and copy bytes to specified location.		*/
	  if ( !WinRTProcessIoBuffer(Session[dwSessionID].Handle, ReadCommand,
   										sizeof(ReadCommand), &ulCount) )
		return ( GetLastError() );

	  *(pulData + i) = ReadCommand[0].value;
    }
  }
 else
  {
   WINRT_CONTROL_ITEM	WriteCommand[] = {{OUTP_LA, 0, 0}};
   for (i=0; i < ulNumElements; i++)
    {
	  WriteCommand[0].port = ulAddress;
     WriteCommand[0].value = *(pulData + i);

			/* Process output command and write byte to specified location.		*/
	  if ( !WinRTProcessIoBuffer(Session[dwSessionID].Handle, WriteCommand,
   										sizeof(WriteCommand), &ulCount) )
		return ( GetLastError() );
    }
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**		DriverControl(): Function to start and stop the WinRT driver. (NT only)	**
**																										**
**		Arguments:  eTask - specifies operation to perform								**
**																										**
**		Return value: dwErrorCode - standard Windows error code for		 			**
**												most recent error		 							**
*******************************************************************************/
DWORD DriverControl( enum DCtrlFunc eTask )
{
 DWORD 			  dwErrorCode = ERROR_SUCCESS;
 SERVICE_STATUS  Status;
 SC_HANDLE 		  hManager, hDriver;

 	/*	Open the System Service Manager.	*/
 hManager = OpenSCManager(NULL, NULL, GENERIC_READ|GENERIC_WRITE|GENERIC_EXECUTE);
 if ( hManager == NULL )
 	return( GetLastError() );

 	/*	Get a handle to the WinRT Driver Service.	*/
 hDriver = OpenService(hManager,"WinRT",GENERIC_READ|GENERIC_WRITE|GENERIC_EXECUTE);
 if ( hDriver == NULL )
  {
 	CloseServiceHandle(hManager);
 	return( GetLastError() );
  }

 	/* Get current status of the driver. */
 if ( !ControlService(hDriver, SERVICE_CONTROL_INTERROGATE, &Status) )
 	dwErrorCode = GetLastError();

/*      Possible service's states on NT.
SERVICE_STOPPED	The service is not running.
SERVICE_START_PENDING	The service is starting.
SERVICE_STOP_PENDING	The service is stopping.
SERVICE_RUNNING	The service is running.
SERVICE_CONTINUE_PENDING	The service continue is pending.
SERVICE_PAUSE_PENDING	The service pause is pending.
SERVICE_PAUSED	The service is paused.
*/

 switch( eTask )
  {
   case ( Start_Driver ):
    		/* Start the WinRT driver if it is not already running.	*/
     if (Status.dwCurrentState == SERVICE_STOPPED)
     	 if ( !StartService(hDriver, 0, NULL) )
       	dwErrorCode = GetLastError();      /* Check if service request failed.	*/
     break;

 	case ( Stop_Driver ):
    		/* Stop the WinRT driver if it is still running.  */
     if (Status.dwCurrentState == SERVICE_RUNNING)
     	 if ( !ControlService(hDriver, SERVICE_CONTROL_STOP, &Status) )
     	   dwErrorCode = GetLastError();
     break;
  }

   /* Close handles to NT Service Manager and WinRT driver.	*/
 CloseServiceHandle(hManager);
 CloseServiceHandle(hDriver);

 return( dwErrorCode );
}


/*******************************************************************************
**   	ConfigureDevice() - This function serves as an interface between			**
**									LabVIEW and the DeviceToRegistry function.         **
**																										**
**		Arguments:  dwDeviceNum - Number of device to configure						**
**						szDeviceID - name of device to configure							**
**						Config - LabVIEW-supplied device data description	         **
**						AtDefault - LabVIEW-supplied default device description.    **
**									  If a configuration item is at its default value	**
**									  that element of the AtDefault array is TRUE, 		**
**									  otherwise, the element is FALSE.						**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD ConfigureDevice(uInt32 dwDeviceNum, CStr szDeviceID, uInt32 *Config,
                        uInt8 *AtDefault)
{
 DEVICE device;							/*	device data description	*/

     /* Initialize all items to not present.                            **
     ** Items which are not present will not be written to              **
     ** the registry, this allows the drivers to apply                  **
     ** appropriate default values for the systems they are running on.	*/

 device.PortAdrs.bItemEmpty      = TRUE;         /* Section 0 Data	*/
 device.PortCnt.bItemEmpty       = TRUE;
 device.MemAdrs.bItemEmpty       = TRUE;
 device.MemCnt.bItemEmpty        = TRUE;
 device.IntNum.bItemEmpty        = TRUE;
 device.IntLevel.bItemEmpty      = TRUE;
 device.IntShare.bItemEmpty      = TRUE;
 device.IntMode.bItemEmpty       = TRUE;
 device.BusNum.bItemEmpty        = TRUE;
 device.BusType.bItemEmpty       = TRUE;
 device.IoIncrem.bItemEmpty      = TRUE;
 device.AllowConflict.bItemEmpty = TRUE;
 device.Disable.bItemEmpty       = TRUE;
 device.PortAdrs1.bItemEmpty     = TRUE;         /* Section 1 Data	*/
 device.PortCnt1.bItemEmpty      = TRUE;
 device.MemAdrs1.bItemEmpty      = TRUE;
 device.MemCnt1.bItemEmpty       = TRUE;

 if ( !(*AtDefault) )
  {
   device.PortAdrs.bItemEmpty = FALSE;        /*  If device item is not at	*/
   device.PortAdrs.dwItemValue = *(Config);   /*  its default value, initialize	*/
  }                                           /*  it properly.					*/

 if ( !(*(AtDefault + 1)) )
  {
   device.PortCnt.bItemEmpty = FALSE;
   device.PortCnt.dwItemValue = *(Config + 1);
  }

 if ( !(*(AtDefault + 2)) )
  {
   device.MemAdrs.bItemEmpty = FALSE;
   device.MemAdrs.dwItemValue = *(Config + 2);
  }

 if ( !(*(AtDefault + 3)) )
  {
   device.MemCnt.bItemEmpty = FALSE;
   device.MemCnt.dwItemValue = *(Config + 3);
  }

 if ( !(*(AtDefault + 4)) )
  {
   device.IntNum.bItemEmpty = FALSE;
   device.IntNum.dwItemValue = *(Config + 4);
  }

 if ( !(*(AtDefault + 5)) )
  {
   device.IntLevel.bItemEmpty = FALSE;
   device.IntLevel.dwItemValue = *(Config + 5);
  }

 if ( !(*(AtDefault + 6)) )
  {
   device.IntShare.bItemEmpty = FALSE;
   device.IntShare.dwItemValue = *(Config + 6);
  }

 if ( !(*(AtDefault + 7)) )
  {
   device.IntMode.bItemEmpty = FALSE;
   device.IntMode.dwItemValue = *(Config + 7);
  }

 if ( !(*(AtDefault + 8)) )
  {
   device.BusNum.bItemEmpty = FALSE;
   device.BusNum.dwItemValue = *(Config + 8);
  }

 if ( !(*(AtDefault + 9)) )
  {
   device.BusType.bItemEmpty = FALSE;
   device.BusType.dwItemValue = *(Config + 9);
  }

 if ( !(*(AtDefault + 10)) )
  {
   device.IoIncrem.bItemEmpty = FALSE;
   device.IoIncrem.dwItemValue = *(Config + 10);
  }

 if ( !(*(AtDefault + 11)) )
  {
   device.AllowConflict.bItemEmpty = FALSE;
   device.AllowConflict.dwItemValue = *(Config + 11);
  }

 if ( !(*(AtDefault + 12)) )
  {
   device.Disable.bItemEmpty = FALSE;
   device.Disable.dwItemValue = *(Config + 12);
  }

 if ( !(*(AtDefault + 13)) )
  {
   device.PortAdrs1.bItemEmpty = FALSE;
   device.PortAdrs1.dwItemValue = *(Config + 13);
  }

 if ( !(*(AtDefault + 14)) )
  {
   device.PortCnt1.bItemEmpty = FALSE;
   device.PortCnt1.dwItemValue = *(Config + 14);
  }

 if ( !(*(AtDefault + 15)) )
  {
   device.MemAdrs1.bItemEmpty = FALSE;
   device.MemAdrs1.dwItemValue = *(Config + 15);
  }

 if ( !(*(AtDefault + 16)) )
  {
   device.MemCnt1.bItemEmpty = FALSE;
   device.MemCnt1.dwItemValue = *(Config + 16);
  }

 lstrcpy( device.szDeviceName, (LPCSTR)(szDeviceID) );
 lstrcpy(device.szVxdName, (LPCSTR)(szDeviceID));    /*  Win95 device name (dummy)	*/
 device.dwDeviceNumber = dwDeviceNum;

 return ( DeviceToRegistry(&device) );
}


/*******************************************************************************
**   	DeviceToRegistry() - Writes device data to Registry. (NT only)          **
**																										**
**		Arguments:  pDevice - device data description                           **
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD DeviceToRegistry(PDEVICE pDevice)
{
 HKEY hkReg;                         	  /* For Registry key	*/
 CHAR szReg[MAX_REGISTRY_NAME];       	  /* Registry name buffer	*/
 DWORD dwDisp;                       	  /* For Registry disposition value	*/
 DWORD dwErrorCode;


 if (lstrlen(pDevice->szDeviceName) == 0)
    return( ERROR_FILE_NOT_FOUND );

     /* Create Registry key for device Section0	*/
 lstrcpy(szReg, DRIVER_REGISTRY_BASE);
 lstrcat(szReg, "\\");
 lstrcat(szReg, pDevice->szDeviceName);
 lstrcat(szReg, "\\Parameters\\Section0");

 dwErrorCode = RegCreateKeyEx(HKEY_LOCAL_MACHINE, szReg, 0, NULL,
                REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS, NULL, &hkReg, &dwDisp);
 if ( dwErrorCode != ERROR_SUCCESS )
    return( dwErrorCode );         /* Catch failure creating Registry key.	*/

  /* Place each device parameter in the Registry (Section 0 params only).	*/
 if ( UpdateRegValue(hkReg, "IoPortAddress",  &(pDevice->PortAdrs))     ||
  	   UpdateRegValue(hkReg, "IoPortCount",    &(pDevice->PortCnt))      ||
      UpdateRegValue(hkReg, "MemoryAddress",  &(pDevice->MemAdrs))      ||
      UpdateRegValue(hkReg, "MemoryCount",    &(pDevice->MemCnt))       ||
      UpdateRegValue(hkReg, "Interrupt",      &(pDevice->IntNum))       ||
      UpdateRegValue(hkReg, "InterruptLevel", &(pDevice->IntLevel))     ||
      UpdateRegValue(hkReg, "InterruptMode",  &(pDevice->IntMode))      ||
      UpdateRegValue(hkReg, "InterruptShare", &(pDevice->IntShare))     ||
      UpdateRegValue(hkReg, "BusNumber",      &(pDevice->BusNum))       ||
      UpdateRegValue(hkReg, "BusType",        &(pDevice->BusType))      ||
      UpdateRegValue(hkReg, "IoIncrement",    &(pDevice->IoIncrem))     ||
      UpdateRegValue(hkReg, "AllowConflicts", &(pDevice->AllowConflict))||
      UpdateRegValue(hkReg, "Disable",        &(pDevice->Disable)) )
    {    													/* Error on Registry Set.	*/
     RegCloseKey(hkReg);            			   /* Close Registry Section 0	*/
     return( GetLastError() );                  /* and return error code.	*/
    }

 RegCloseKey(hkReg);                 /* Close Registry Section 0	*/

 	/* If any parameters for a second section of I/O for this device, these  **
   ** are saved in "Section1" key under "Parameters".                       **
   ** Create Registry key for new device Section1 - even if no values.	    */

 lstrcpy(szReg, DRIVER_REGISTRY_BASE);
 lstrcat(szReg, "\\");
 lstrcat(szReg, pDevice->szDeviceName);
 lstrcat(szReg, "\\Parameters\\Section1");

 dwErrorCode = RegCreateKeyEx(HKEY_LOCAL_MACHINE, szReg, 0, NULL,
               REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS, NULL, &hkReg, &dwDisp);
 if ( dwErrorCode != ERROR_SUCCESS )
  	 return( dwErrorCode );         /* Catch failure creating Registry key.	*/

  	/* Enter any parameters for Section1	*/
 if ( UpdateRegValue(hkReg, "IoPortAddress",  &(pDevice->PortAdrs1)) ||
   	UpdateRegValue(hkReg, "IoPortCount",    &(pDevice->PortCnt1))  ||
     	UpdateRegValue(hkReg, "MemoryAddress",  &(pDevice->MemAdrs1))  ||
      UpdateRegValue(hkReg, "MemoryCount",    &(pDevice->MemCnt1)) )
 	 {     											  	/* Error on Registry Set  */
     RegCloseKey(hkReg);             	  		/* Close Reg Section 1,   */
     return( GetLastError() );           		/* get error and return	  */
 	 }

 RegCloseKey(hkReg);                 	  /* Close Reg Section 1	 */
 return( ERROR_SUCCESS );                /* Everything OK	*/
}


/*******************************************************************************
**   UpdateRegValue() - Updates individual Registry entry.							**
**   Arguments:  hKey     - open registry key										   **
**            	  pszName  - name of key value to set									**
**            	  pItem    - Item structure to set                             **
**                                                                            **
**		Return value:	Success = FALSE														**
**							Failure = TRUE															**
*******************************************************************************/
DWORD UpdateRegValue(HKEY hKey, PCHAR pszName, PDEVICE_ITEM pItem)
{
 DWORD dwErrorCode;
 DWORD dwValue, dwLength;

	/* If item is empty, verify current Registry value. */
 if ( pItem->bItemEmpty )
   {
	 dwLength = sizeof(dwValue);
    if ( RegQueryValueEx(hKey, pszName, 0, NULL, (LPBYTE)&dwValue, &dwLength) )
 	 	return( FALSE );          /* Value is at default, do not overwrite. */
    else
	   dwErrorCode = RegDeleteValue(hKey, pszName);
   }

 else      /* If item is not empty, write new value to Registry.	*/
   {
 	 dwErrorCode = RegSetValueEx(hKey, pszName, 0, REG_DWORD,
	  				(BYTE *CONST)(&pItem->dwItemValue), sizeof(DWORD));
   }

 return( dwErrorCode != ERROR_SUCCESS );
}


/*******************************************************************************
**   DeleteDevice() - Deletes ALL information corresponding to the         	**
**								  specified device from the Registry.						**
**																										**
**		Arguments:	szDeviceID - identifies I/O device to control					**
**                                                                            **
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD DeleteDevice(CStr szDeviceID)
{
 HKEY hkReg;                         	  		/* For Registry key	*/
 CHAR szReg[MAX_REGISTRY_NAME];            /* Registry name buffer	*/
 DWORD dwErrorCode;

 lstrcpy(szReg, DRIVER_REGISTRY_BASE);
 lstrcat(szReg, "\\");
 lstrcat(szReg, (LPCSTR)(szDeviceID) );
 lstrcat(szReg, "\\Parameters\\");

/* Windows NT doesn't allow us to delete a Key with subkeys.  Therefore, we  	**
** must first delete the subkeys and work our way up in the Reg hierarchy 		**
** until we reach the top-level Key.													   */

 dwErrorCode = RegOpenKeyEx(HKEY_LOCAL_MACHINE, szReg, 0,KEY_ALL_ACCESS, &hkReg);
 if ( dwErrorCode != ERROR_SUCCESS )
    return( dwErrorCode );         /* Catch failure opening Registry key.	*/

 dwErrorCode = RegDeleteKey( hkReg,"Section0" );
 if ( dwErrorCode != ERROR_SUCCESS )
    return( dwErrorCode );         /* Catch failure deleting Registry key.	*/

 dwErrorCode = RegDeleteKey( hkReg,"Section1" );
 if ( dwErrorCode != ERROR_SUCCESS )
    return( dwErrorCode );

 RegCloseKey(hkReg);                 /* Finished accessing Parameters key.	*/

 lstrcpy(szReg, DRIVER_REGISTRY_BASE);
 lstrcat(szReg, "\\");
 lstrcat(szReg, (LPCSTR)(szDeviceID) );

 dwErrorCode = RegOpenKeyEx(HKEY_LOCAL_MACHINE, szReg, 0,KEY_ALL_ACCESS, &hkReg);
 if ( dwErrorCode != ERROR_SUCCESS )
    return( dwErrorCode );         /* Catch failure opening Registry key.	*/

 dwErrorCode = RegDeleteKey( hkReg,"Parameters" );
 if ( dwErrorCode != ERROR_SUCCESS )
    return( dwErrorCode );         /* Catch failure deleting Registry key.	*/

 RegCloseKey(hkReg);                 /* Finished accessing Device key.	*/

	/* Last step, delete Device Key.	*/
 dwErrorCode = RegOpenKeyEx(HKEY_LOCAL_MACHINE, DRIVER_REGISTRY_BASE,
 									 0,KEY_ALL_ACCESS, &hkReg);
 if ( dwErrorCode != ERROR_SUCCESS )
    return( dwErrorCode );         /* Catch failure opening Registry key.	*/

 dwErrorCode = RegDeleteKey( hkReg,(LPCSTR)(szDeviceID) );
 RegCloseKey(hkReg);                 /* Finished accessing Registry.	*/

 return( dwErrorCode );         /* Catch failure deleting Registry key.	*/
}


/*******************************************************************************
**		GetSessionConfiguration():	This function returns the current Registry	**
**									settings for the specified session.						**
**																										**
**		Arguments:	dwSessionID - number that describes I/O session					**
**						Config - returns array with session configuration values	   **
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
GetSessionConfiguration(DWORD dwSessionID, uInt32 *Config)
{
 WINRT_CONFIGURATION	Settings;
 DWORD dwLength;

 if ( !WinRTGetConfiguration(Session[dwSessionID].Handle, &Settings, &dwLength) )
 	return ( GetLastError() );

 *(Config) = Settings.portMin;                   		/* Section 0 Data	*/

 if ( Settings.portMax )// != Settings.portMin )
 	*(Config + 1) = (Settings.portMax - Settings.portMin) + 1;
// else
// 	*(Config + 1) = 0;

 *(Config + 2) = Settings.memMin;

 if ( Settings.memMax )//!= Settings.memMin )
 	*(Config + 3) = (Settings.memMax - Settings.memMin) + 1;
// else
// 	*(Config + 3) = 0;

 if ( Settings.interruptVector > 50 )          	/* Catch "undefined" numbers.	*/
   *(Config + 4) = 0;                           /* It generally indicates no	*/
 else                                           /* interrupts were assigned 	*/
 	*(Config + 4) = Settings.interruptVector;    /* to this device.				*/

 if ( Settings.interruptLevel > 50 )
   *(Config + 5) = 0;
 else
 	*(Config + 5) = Settings.interruptLevel;

 *(Config + 8) = Settings.busNumber;
 *(Config + 9) = Settings.busType;
 *(Config + 10) = Settings.ioIncrement;

 *(Config + 13) = Settings.portMin1;                   		/* Section 1 Data	*/

 if ( Settings.portMax1 )//!= Settings.portMin1 )
 	*(Config + 14) = (Settings.portMax1 - Settings.portMin1) + 1;
// else
// 	*(Config + 14) = 0;

 *(Config + 15) = Settings.memMin1;

 if ( Settings.memMax1 )//!= Settings.memMin1 )
   *(Config + 16) = (Settings.memMax1 - Settings.memMin1) + 1;
// else
// 	*(Config + 16) = 0;

 return ( ERROR_SUCCESS );
}


/*******************************************************************************
**	Search():	Used by AllocateNewSession to find available device.				**
**     			Searches for Key in List.													**
**	Arguments:	Key - number to search for													**
**					List - array to be searched for value									**
**					nElements - number of elements in List									**
**																										**
**	Returns:	Success - index of Key in List												**
**				Failure - -1                                                      **
*******************************************************************************/
int Search(int iKey, int List[], int nElements)
{
 register i = 0;

 while ( (List[i] != iKey) && ( i < nElements) ) i++;

 if (List[i] != iKey)
 	return(-1);

 return(i);
}


/*******************************************************************************
**	Delay():	This function waits for the specified number of milliseconds		**
**     		and then returns to its caller.												**
**	Arguments:	dwMilliseconds - number of milliseconds to wait						**
**																										**
**	Returns:	none																					**
*******************************************************************************/
void Delay(DWORD dwMilliseconds)
{
 DWORD start;

 start = GetTickCount();
 while ( (GetTickCount() - start) < dwMilliseconds );

 return;
}


/*******************************************************************************
**	   MemControl():	Function to open and close memory maps for any session.	**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						eTask - type of operation to perform								**
**						BaseMem1 - base address of primary memory block				   **
**						CountMem1 - length of primary memory block			   		**
**						BaseMem2 - base address of secondary memory block				**
**						CountMem2 - length of secondary memory block			   		**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD MemControl(DWORD dwSessionID, enum MemCommand eTask, DWORD ulBaseMem1,
						DWORD ulCountMem1, DWORD ulBaseMem2, DWORD ulCountMem2)
{
 ULONG	 		   ulCount;
 WINRT_MEMORY_MAP Region;        		/* Description of memory block.  */

 switch( eTask )
  {
   case ( MapMemory ):
    	{  /* If session has memory resources, map them.	*/
    	 if ( ulCountMem1 > 0 )
     		{
 			 Region.address = (PVOID)(ulBaseMem1);
 			 Region.length  = ulCountMem1;

   			/* Map in primary memory range.  */
          if ( !WinRTMapMemory(Session[dwSessionID].Handle, &Region,
      								&Session[dwSessionID].PhysMem1, &ulCount) )
          	return( GetLastError() );
     		}

    	 if ( ulCountMem2 > 0 )
     		{
 			 Region.address = (PVOID)(ulBaseMem2);
 			 Region.length  = ulCountMem2;

   			/* Map in secondary memory range.  */
 			 if ( !WinRTMapMemory(Session[dwSessionID].Handle, &Region,
      								&Session[dwSessionID].PhysMem2, &ulCount) )
          	return( GetLastError() );
     		}
    	 break;
      }

   case ( UnmapMemory ):
   	{	/* Unmap memory allocated to the session.	*/
    	 if ( ulCountMem1 > 0 )
     		{	/* Unmap primary memory range.  */
	 		 if ( !WinRTUnMapMemory(Session[dwSessionID].Handle,
      								Session[dwSessionID].PhysMem1, &ulCount) )
          	return( GetLastError() );

          	/*	Invalidate memory pointer.	*/
      	 Session[dwSessionID].PhysMem1 = NULL;
     		}

    	 if ( ulCountMem2 > 0 )
     		{	/* Unmap secondary memory range.  */
	 		 if ( !WinRTUnMapMemory(Session[dwSessionID].Handle,
      								Session[dwSessionID].PhysMem2, &ulCount) )
          	return( GetLastError() );

            /*	Invalidate memory pointer.	*/
      	 Session[dwSessionID].PhysMem2 = NULL;
         }
    	 break;
      }

   default: break;
  }

 return( ERROR_SUCCESS );
}


/*******************************************************************************
**	   WaitForInterrupt():	Function to asynchronuosly wait for an interrupt.	**
**																										**
**		Arguments:  dwSessionID - number that describes I/O session					**
**						iOccurrence - LabVIEW occurrence to be set						**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD WaitForInterrupt(DWORD dwSessionID, int32 iOccurrence)
{
 HANDLE hSessionDevice;						  /* Handle to session's WinRT device.	*/
 HANDLE hIsrThread;						  	  /* Handle to new IsrFunction thread.	*/
 DWORD  dwThreadId;        				  /* ID of IsrFunction thread.			*/
 char szDeviceName[MAX_WINRT_NAME];
 char dummy[MAX_DEVICE_DIGITS];

 	/* Check for IsrThread initialization.	*/
 if (Session[dwSessionID].IsrEvent == NULL )
 	{
		/* Make shure that new thread will not be terminated.	*/
    Session[dwSessionID].TerminateIsr = FALSE;

		/* Create thread synchronization object.  The event object is set up		*/
      /* as an auto-reset event initialized to a nonsignaled state.				*/
    Session[dwSessionID].IsrEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (Session[dwSessionID].IsrEvent == NULL)
    	 return( GetLastError() );

	 	/*	Generate DeviceName string with session device number. */
	 lstrcpy(szDeviceName, "\\\\.\\WinRTdev");
 	 itoa((int)(Session[dwSessionID].Device), dummy, 10);
	 lstrcat(szDeviceName, dummy);

	 	/*	Obtain a new handle to session's device (for the new thread).	*/
	 hSessionDevice = CreateFile( (LPCTSTR)(szDeviceName), 0,
 						FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);

		/* Catch invalid handle.	*/
	 if ( hSessionDevice == INVALID_HANDLE_VALUE )
    	{
       CloseHandle(Session[dwSessionID].IsrEvent);
       Session[dwSessionID].IsrEvent = NULL;
	  	 return( GetLastError() );
      }

	   /* Save interrupt settings in Session Table structure.	*/
	 Session[dwSessionID].Occurrence = iOccurrence;
    ID = dwSessionID;

	   /* Create IsrFunction thread in suspended state.	*/
	 hIsrThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)IsrFunction,
										hSessionDevice, CREATE_SUSPENDED, &dwThreadId);

	   /* Check thread creation.	*/
	 if ( hIsrThread == NULL )
    	{
       CloseHandle(hSessionDevice);            	  /* Failed to create thread.	*/
       CloseHandle(Session[dwSessionID].IsrEvent);
       Session[dwSessionID].IsrEvent = NULL;
       return ( GetLastError() );
      }

	   /* Set the priority for the new thread.	*/
	 SetThreadPriority(hIsrThread, THREAD_PRIORITY_TIME_CRITICAL);

	    /* Get a handle to the LabVIEW window to be notified of interrupt.	*/
	 Session[dwSessionID].WndHandle = FindWindow(NULL,(LPCTSTR)("LabVIEW"));
	 if ( Session[dwSessionID].WndHandle == NULL )
    	{
       Session[ID].TerminateIsr = TRUE;
       ResumeThread(hIsrThread);
       CloseHandle(hIsrThread);
	 	 return( GetLastError() );
      }

	   /* Start up IsrFunction thread and close setup handle.	*/
    ResumeThread(hIsrThread);
    CloseHandle(hIsrThread);
   }

 else
 	{	/* If thread has been initialized, wake it up to service interrupts.	*/
    ID = dwSessionID;
    SetEvent(Session[dwSessionID].IsrEvent);
   }

 return(GetLastError());
}


/*******************************************************************************
**	   IsrFunction():	Thread to service session interrupts.							**
**																										**
**		Arguments:  hSessionDevice - handle to session's WinRT device				**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD IsrFunction(HANDLE hSessionDevice)
{
 DWORD dwLength;
 WINRT_CONTROL_ITEM CommandBuffer[2];

   /*	Set up command buffer with universal interrupt settings.	 */
	/* Driver will ALWAYS indentify the interrupt as ours and	 */
   /* will NOT CLEAR the interrupt upon its occurrence.			 */
 CommandBuffer[0].command = INTRP_ID_ALWAYS;
 CommandBuffer[1].command = INTRP_CLEAR_NOP;

Loop:
 if (Session[ID].TerminateIsr)
  {
   CloseHandle(hSessionDevice);
   CloseHandle(Session[ID].IsrEvent);
   Session[ID].IsrEvent = NULL;
   Session[ID].WndHandle = NULL;
 	ExitThread(ERROR_SUCCESS);
  }
 else
  {	 /* Make thread wait for interrupt occurrence.	*/
	WinRTWaitForInterrupt(hSessionDevice, CommandBuffer, sizeof(CommandBuffer), &dwLength);

	    /* Notify LabVIEW of interrupt.	*/
	SendMessage(Session[ID].WndHandle, WM_USER+2, 0, Session[ID].Occurrence);

//   if (Session[ID].IsrEvent == NULL)  		Debugging code...
//   	Beep(500, 500);

    	 /* Suspend thread execution.	*/
	if (WaitForSingleObject(Session[ID].IsrEvent, INFINITE) == WAIT_FAILED)
   	Session[ID].TerminateIsr = TRUE;
   goto Loop;
  }

 return (ERROR_SUCCESS);
}


/*******************************************************************************
**   	ConfigurePCIDevice() - This function writes a PCI device's 					**
**									information into the Registry.         				**
**																										**
**		Arguments:  szDeviceID - name of device to configure							**
**						dwVendorId - PCI vendor Id									         **
**						dwCardId - Card's PCI device Id										**
**																										**
**		Return value:	Success = 0																**
**							Failure = standard Windows error code 							**
**										 (defined in winerror.h)								**
*******************************************************************************/
DWORD ConfigurePCIDevice(CStr szDeviceID,	uInt32 dwVendorId, uInt32 dwCardId)
{
 HKEY hkReg;                         	  /* For Registry key	*/
 CHAR szReg[MAX_REGISTRY_NAME];       	  /* Registry name buffer	*/
 DWORD dwDisp;                       	  /* For Registry disposition value	*/
 DWORD dwErrorCode;

     /* Create Registry key for device Section0	*/
 lstrcpy(szReg, DRIVER_REGISTRY_BASE);
 lstrcat(szReg, "\\");
 lstrcat(szReg, (LPCSTR)(szDeviceID) );
 lstrcat(szReg, "\\Parameters\\Section0");

 dwErrorCode = RegCreateKeyEx(HKEY_LOCAL_MACHINE, szReg, 0, NULL,
                REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS, NULL, &hkReg, &dwDisp);
 if ( dwErrorCode != ERROR_SUCCESS )
    return( dwErrorCode );         /* Catch failure creating Registry key.	*/

	/* Write VendorId value to Registry.	*/
 dwErrorCode = RegSetValueEx(hkReg, "VendorId", 0, REG_DWORD,
 							(BYTE *CONST)(&dwVendorId), sizeof(dwVendorId));

	/* Write CardId value to Registry.	*/
 dwErrorCode = RegSetValueEx(hkReg, "CardId", 0, REG_DWORD,
 							(BYTE *CONST)(&dwCardId), sizeof(dwCardId));

 RegCloseKey(hkReg);               /* Close Registry Section 0	*/

 return( dwErrorCode );            /* Return outcome of Registry calls	*/
}

